#!/usr/bin/env python3

import argparse
import re
import math

import numpy as np
from PIL import Image

parser = argparse.ArgumentParser()
parser.add_argument('--output', '-o', help='output base name (witout .gfx .map. pal)', default=None)
parser.add_argument('--verbose', '-v', help='verbose instrumentation spam', action="store_true")
parser.add_argument('--full_map', '-f', help='produce a 256x256 map regardless of original image dimensions, pads image with empty space', action="store_true")
parser.add_argument('--map_delta', '-d', help='add this number to the tile index in the map, e.g., for loading an image after sprites in BANK1', type=int, default=0)
parser.add_argument('--pal', '-p', help='set palette in tiles to this number, for special cases in code (consider something more robust instead?)', type=int, default=-1)
parser.add_argument('files', nargs=argparse.REMAINDER, help='list of input files')
# parser.add_argument('files', help='list of input files')


args = parser.parse_args()
files = args.files
print(f"reading {files}")
if not args.output:
    args.output =  re.sub(r"\..*", "", files[0])
print(f"writing to {args.output}")
# print("NOTE: it's best if all the images use the same palette or else you'll have to load multiple palettes. They'll be indexed in the map data.")

map_delta = args.map_delta
pal_override = args.pal


# TODO write out correct .inc and .h files for multi-frame combined banks automatically?
# finc = open(f"{args.output}.inc", "wt")
# fh = open(f"{args.output}.h", "wt")


class map_entry:
    tile = 0
    palette = 0
    hflip = 0
    vflip = 0
    def __init__(self, tile, palette):
        self.tile = tile
        self.palette = palette
        if pal_override != -1:
            self.palette = pal_override


palettes = []
tiles = [] # one list of np arrays of 8x8 tiles 
maps = [] # list of lists of map_entry

reused = 0

for fn in files:
    i = Image.open(fn)
    # print(i, i.info)
    colors = len([c for c in i.histogram() if c > 0])
    if colors > 16:
        print(f"{fn} has {colors} colors, converting. Consider using an external tool for this to verify results before converting to .gfx")
        if i.mode == 'P':
            i = i.convert() # stupid workaround to make quantization and dithering work when starting with palette-based image
        i = i.convert(mode='P', colors=16)
        if not 'transparency' in i.info:
            i.info['transparency'] = 15
        print(i.getpalette())
    a = np.array(i)

    pal = np.array(i.getpalette())
    palette_index = 0
    # see whether we're using the same palettes
    match = False
    for pali in range(len(palettes)):
        if np.array_equal(palettes[pali], pal):
            palette_index = pali
            match = True
            break
    if not match:
        palettes.append(pal)
        palette_index = len(palettes)-1
    tilemap = []
    if args.verbose: print(palette_index, palettes)    
    transparent_tile = np.zeros((8,8), dtype=np.uint8)
    if 'transparency' in i.info:
        image_transparency = i.info['transparency']
    else:
        print(f"warning, assuming transparent color 0 for file with no transparent pixels {fn}")
        image_transparency = 0
    # remap transparency so that it is 0
    if image_transparency != 0:
        print("remapping transparent pixels")
        # move color out of slot 0
        c0 = palettes[palette_index][image_transparency*3]
        c1 = palettes[palette_index][image_transparency*3+1]
        c2 = palettes[palette_index][image_transparency*3+2]
        palettes[palette_index][image_transparency*3] = palettes[palette_index][0]
        palettes[palette_index][image_transparency*3+1] = palettes[palette_index][1]
        palettes[palette_index][image_transparency*3+2] = palettes[palette_index][2]
        palettes[palette_index][0] = c0
        palettes[palette_index][1] = c1
        palettes[palette_index][2] = c2
        for y in range(a.shape[0]):
            for x in range(a.shape[1]):
                if a[y,x] == 0:
                    a[y,x] = image_transparency
                elif a[y,x] == image_transparency:
                    a[y,x] = 0
    if args.full_map:
        MAX_X = 32
        MAX_Y = 32
    else:
        MAX_X = math.ceil(i.size[0] / 8)
        MAX_Y = math.ceil(i.size[1] / 8)
    for y in range(MAX_Y):
        for x in range(MAX_X):
            # if args.verbose: print(f"{fn} {x}x{y}")
            tile = a[y*8:y*8+8,x*8:x*8+8]
            # print(f"{x}x{y} {tile}")
            if tile.size < 64:
                if args.verbose: print(f"{fn} {x}x{y}: expanding tile")
                oldtile = tile
                tile = np.zeros((8,8), dtype=np.uint8)
                tile[0:oldtile.shape[0],0:oldtile.shape[1]] = oldtile
            # find tile in tiles[] or add to tiles[], put index into maps[]
            # TODO check for hflip and vflip matches
            match = False
            for t in range(len(tiles)):
                if np.array_equal(tile, tiles[t]): # duplicate!
                    tilemap.append(map_entry(t, palette_index))
                    match = True
                    reused += 1
                    # if args.verbose: print(f"reusing tile {t}, {reused} reused")
                    # print(f"{tile} vs {tiles[t]}")
                    break
            if not match:
                # if args.verbose: print(f"adding new tile {len(tiles)} {x}x{y}")
                # print(tile)
                tiles.append(tile)
                if (len(tiles) > 512):
                    print("Out of bank space :(")
                    # exit(1)
                tilemap.append(map_entry(len(tiles)-1, palette_index))
    maps.append(tilemap)
    # print(tilemap)
    # print(tilemap[0])
    print(f"{reused} tiles reused in {fn}")


for i in range(len(palettes)):
    if len(files) > 1:
        fpal = open(f"{args.output}_{i}.pal", "wb");
        # finc.write(f'global _{args.output}_{i}_pal')
        # finc.write(f'_{args.output}_{i}_pal: incbin"{args.output}_{i}.pal"\n')
        # fh.write(f'{args.output}_{i}')
    else:
        fpal = open(f"{args.output}.pal", "wb")
    for cnum in range(16):
        c = ((palettes[i][cnum*3] >> 4) << 8) + ((palettes[i][cnum*3+1] >> 4) << 4) + (palettes[i][cnum*3+2] >> 4)
        fpal.write(int(c).to_bytes(2,'little'))
    fpal.close()

fgfx = open(f"{args.output}.gfx", "wb")
for tile in tiles[:512]:
    for row in tile:
        packed = np.zeros(len(row)>>1, dtype=np.uint8) # 4 really, tiles are always 8x8
        for i in range(len(row)>>1):
            packed[i] = ((row[i*2] & 0xf) << 4) + (row[i*2 + 1] & 0xf) # slow
        packed.tofile(fgfx)
fgfx.close()

for i in range(len(maps)):
    map = maps[i]
    if len(maps) > 1:
        fm = open(f"{args.output}_{i}.map", "wb")
    else:
        fm = open(f"{args.output}.map","wb")
    for me in map:
        i = ((me.tile + map_delta) & 0x1ff) + (me.palette << 9) + (me.hflip << 14) + (me.vflip << 15)
        fm.write(i.to_bytes(2, 'little'))
    fm.close()
